from abc import ABC, abstractmethod
import sys
import subprocess
import os.path
from enum import Enum, auto


def sanitize_path(path):
    return os.path.normpath(path)


def sanitize_paths(paths):
    ret = []
    for p in paths:
        p = sanitize_path(p)
        if len(p) != 0:
            ret.append(p)
    return ret


def build_target_key(target_name: str, toolchain_name: str, build_type: str):
    return "{}_{}_{}".format(target_name, toolchain_name, build_type)


class TargetKind(Enum):
    STATIC_LIB = auto()
    DYNAMIC_LIB = auto()
    EXECUTABLE = auto()

    @staticmethod
    def from_string(s: str):
        return {
            "rapid_static_lib": TargetKind.STATIC_LIB,
            "rapid_dynamic_lib": TargetKind.DYNAMIC_LIB,
            "rapid_exe": TargetKind.EXECUTABLE,
        }[s]

    def to_string(self):
        return {
            TargetKind.STATIC_LIB: "rapid_static_lib",
            TargetKind.DYNAMIC_LIB: "rapid_dynamic_lib",
            TargetKind.EXECUTABLE: "rapid_exe",
        }[self.value]


BUILD_TYPE_DEBUG = "debug"
BUILD_TYPE_RELEASE = "release"

TOOLCHAIN_MSVC16 = "msvc16"
TOOLCHAIN_LLVM13 = "llvm13"


class Logger(ABC):
    @abstractmethod
    def log_info(self, *args):
        pass

    @abstractmethod
    def log_error(self, *args):
        pass


class LoggerDefault(Logger):
    def log_info(self, *args):
        print(*args, file=sys.stdout)

    def log_error(self, *args):
        print(*args, file=sys.stderr)


class ProcessRunner(ABC):
    @abstractmethod
    def run(self, args, *, env, check):
        pass


class ProcessRunnerDefault(ProcessRunner):
    def run(self, args, *, env, check):
        subprocess.run(args, env=env, check=check)


class SystemConfigManifestProvider(ABC):
    @abstractmethod
    def run(self, working_dir_abs_path):
        pass

    @abstractmethod
    def get_toolchains_manifest(self):
        pass

    @abstractmethod
    def get_third_party_manifest(self):
        pass


class ToolchainSettingsProvider(ABC):
    @abstractmethod
    def get_toolchain_settings(self):
        pass


class BuildScriptEmitter(ABC):
    @abstractmethod
    def filename(self):
        pass

    @abstractmethod
    def contents(self, build_dir_abs_path, toolchains_manifest, toolchains_settings, build_types, targets_names,
                 targets_impls) -> str:
        pass



from __future__ import annotations

from typing import Any
import importlib.resources
import fs
import os.path
import json

from . import base
from . import conan_generator
from .data import conan_profiles


def to_rapid_os(os_name: str) -> str | None:
    if os_name == "Windows":
        return "windows"
    return None


def to_rapid_toolchain(conan_compiler: str, version: str) -> str | None:
    if conan_compiler == "Visual Studio":
        if version == "16":
            return base.TOOLCHAIN_MSVC16
    if conan_compiler == "clang":
        if version == "13":
            return base.TOOLCHAIN_LLVM13
    return None


def _generate_toolchain_msvc_from_conan_vc_vars(vc_vars):
    toolchain = {}

    generate_sdk_paths = False
    if generate_sdk_paths:
        windows_sdk_dir = base.sanitize_path(vc_vars["WindowsSdkDir"])
        windows_sdk_version = vc_vars["WindowsSDKVersion"]

        toolchain["windows_sdk_include_path"] = os.path.join(f"{windows_sdk_dir}Include", f"{windows_sdk_version}")
        toolchain["windows_sdk_lib_path"] = os.path.join(f"{windows_sdk_dir}Lib", f"{windows_sdk_version}")

    vc_tools_install_dir = base.sanitize_path(vc_vars["VCToolsInstallDir"])
    vc_host_arch = vc_vars["VSCMD_ARG_HOST_ARCH"]
    vc_target_arch = vc_vars["VSCMD_ARG_TGT_ARCH"]

    vc_bin_path = os.path.join(f"{vc_tools_install_dir}", "bin", f"Host{vc_host_arch}", f"{vc_target_arch}")
    # vc_lib_path = os.path.join(f"{vc_tools_install_dir}", "lib", f"{vc_target_arch}")
    # vc_include_path = base.sanitize_path(vc_tools_install_dir)

    toolchain["bin_path"] = vc_bin_path
    toolchain["compiler_path"] = os.path.join(vc_bin_path, "cl.exe")
    toolchain["compiler_family"] = "msvc"
    toolchain["linker_path"] = os.path.join(vc_bin_path, "link.exe")
    toolchain["librarian_path"] = os.path.join(vc_bin_path, "lib.exe")

    msvc_extra_files = [
        'c1.dll',
        'c1xx.dll',
        'c2.dll',
        'atlprov.dll',  # Only needed if using ATL
        'msobj140.dll',
        'mspdb140.dll',
        'mspdbcore.dll',
        'mspdbsrv.exe',
        'mspft140.dll',
        'msvcp140.dll',
        'tbbmalloc.dll',  # Required as of 16.2(14.22 .27905)
        'vcruntime140.dll',
        os.path.join('1033', 'clui.dll'),
        os.path.join('1033', 'mspft140ui.dll'),  # Localized messages for static analysis
    ]

    msvc_extra_files = [os.path.join(vc_bin_path, p) for p in msvc_extra_files if len(p) != 0]

    toolchain["compiler_extra_files"] = base.sanitize_paths(msvc_extra_files)
    toolchain["toolchain_include_dirs"] = base.sanitize_paths(vc_vars["INCLUDE"])
    toolchain["toolchain_lib_dirs"] = base.sanitize_paths(vc_vars["LIB"])
    return toolchain


def _generate_toolchain_llvm_from_conan_vc_vars(vc_vars):
    use_vs_embedded_llvm = False
    if use_vs_embedded_llvm:
        llvm_bin_path = os.path.join(base.sanitize_path(vc_vars["VSINSTALLDIR"]), "VC", "Tools", "Llvm", "x64", "bin")
    else:
        llvm_bin_path = os.path.join("C:\\", "Program Files", "LLVM", "bin")

    toolchain = {
        "compiler_path": os.path.join(llvm_bin_path, "clang++.exe"),
        "compiler_family": "clang",
        "linker_path": os.path.join(llvm_bin_path, "lld-link.exe"),
        "librarian_path": os.path.join(llvm_bin_path, "llvm-ar.exe"),
        "compiler_extra_files": [],
        "toolchain_include_dirs": base.sanitize_paths(vc_vars["INCLUDE"]),
        "toolchain_lib_dirs": base.sanitize_paths(vc_vars["LIB"]),
    }
    return toolchain


def get_toolchain_manifest_filename():
    return "toolchain_gen.json"


def generate_toolchain_manifest_from_conan_vc_vars(toolchain_name, vc_vars):
    if toolchain_name == base.TOOLCHAIN_MSVC16:
        toolchain = _generate_toolchain_msvc_from_conan_vc_vars(vc_vars)
    elif toolchain_name == base.TOOLCHAIN_LLVM13:
        toolchain = _generate_toolchain_llvm_from_conan_vc_vars(vc_vars)
    else:
        raise NotImplementedError("Toolchain '{}' is not implemented.".format(toolchain_name))

    toolchain_manifest = {
        toolchain_name: toolchain
    }

    return json.dumps(toolchain_manifest, indent=4)


def get_third_party_manifest_filename():
    return "third_party_gen.json"


def generate_third_party_manifest_from_conan_deps_info(deps_cpp_info):
    targets_interfaces = {}
    for dep in deps_cpp_info.deps:
        dep_cpp_info = deps_cpp_info[dep]

        dep_interface = {
            "kind": base.TargetKind.STATIC_LIB.to_string(),  # TODO: determine if lib static or dynamic
            "include_dirs": base.sanitize_paths(dep_cpp_info.include_paths),
            "definitions": {p: None for p in dep_cpp_info.defines},  # TODO: parse defines
            "link_libs": [],
            "link_libs_external": base.sanitize_paths(dep_cpp_info.libs),
            "link_libs_external_dirs": base.sanitize_paths(dep_cpp_info.lib_paths),
            "runtime_libs": [],
            "_sharedlinkflags": dep_cpp_info.sharedlinkflags,
            "_exelinkflags": dep_cpp_info.exelinkflags,
            "_system_libs": dep_cpp_info.system_libs,
            "_cxxflags": dep_cpp_info.cxxflags,
        }
        targets_interfaces[dep] = dep_interface
    return json.dumps(targets_interfaces, indent=4)


class SystemConfigManifestProviderConan(base.SystemConfigManifestProvider):
    def __init__(self, *, logger, process_runner, conanfile_path, execute_conan, base_to_conan_build_type):
        self.logger = logger
        self.process_runner = process_runner
        self.conanfile_path = conanfile_path
        self.execute_conan = execute_conan
        self.base_to_conan_build_type = base_to_conan_build_type

        self.third_party_manifest: dict[str, Any] = {}
        self.toolchains_manifest: dict[str, Any] = {}

    @staticmethod
    def install_combination_to_install_dir_name(profile_name: str, build_type: str) -> str:
        return f"{profile_name}-{build_type}"

    def run(self, working_dir_abs_path):
        supported_conan_build_types = [
            "Debug",
            "Release"
        ]

        for conan_build_type in self.base_to_conan_build_type.values():
            if conan_build_type not in supported_conan_build_types:
                raise RuntimeError(f"Unsupported conan build type '{conan_build_type}'")

        conan_export_invocation = []
        conan_export_resource = ""

        for resource in importlib.resources.contents(conan_generator):
            if resource == "conanfile.py":
                with importlib.resources.path(conan_generator, resource) as conan_generator_path:
                    conan_export_resource = resource
                    conan_export_invocation = ["conan", "export", f"{conan_generator_path}"]
                break

        if conan_export_resource == "":
            raise RuntimeError("Internal error: unable to find Conan generator resource.")

        conan_install_invocations = []
        conan_install_invocations_resources = []
        conan_install_combinations = []

        for resource in importlib.resources.contents(conan_profiles):
            profile_name, ext = fs.path.splitext(resource)
            if ext == ".txt":
                for build_type, conan_build_type in self.base_to_conan_build_type.items():

                    install_dir_name = self.install_combination_to_install_dir_name(profile_name, build_type)
                    install_dir_abs_path = fs.path.join(working_dir_abs_path, install_dir_name)

                    with importlib.resources.path(conan_profiles, resource) as profle_path:
                        argv = [
                            "conan", "install",
                            f"--profile={profle_path}",
                            "-s", f"build_type={conan_build_type}",
                            f"--install-folder={install_dir_abs_path}",
                            "--build=missing",
                            f"{self.conanfile_path}"
                        ]

                    conan_install_invocations.append(argv)
                    conan_install_invocations_resources.append(resource)
                    conan_install_combinations.append((profile_name, build_type))

        if self.execute_conan:
            abs_conan_home_path = fs.path.join(working_dir_abs_path, "conan_home")

            env = os.environ.copy()
            env["CONAN_USER_HOME"] = abs_conan_home_path

            with importlib.resources.path(conan_generator, conan_export_resource) as conan_generator_path:
                self.logger.log_info("Invoking: {}".format(" ".join(conan_export_invocation)))
                self.process_runner.run(conan_export_invocation, env=env, check=True)

            for args, resource in zip(conan_install_invocations, conan_install_invocations_resources):
                self.logger.log_info("Invoking: {}".format(" ".join(argv)))
                with importlib.resources.path(conan_profiles, resource) as _:
                    self.process_runner.run(args, env=env, check=True)

        with fs.open_fs(working_dir_abs_path) as work_fs:
            for profile_name, build_type in conan_install_combinations:
                install_dir_name = self.install_combination_to_install_dir_name(profile_name, build_type)

                path = fs.path.join(install_dir_name, get_toolchain_manifest_filename())
                if not work_fs.exists(path):
                    raise RuntimeError(f"Expected file '{path}' does not exist.")

                path = fs.path.join(install_dir_name, get_third_party_manifest_filename())
                if not work_fs.exists(path):
                    raise RuntimeError(f"Expected file '{path}' does not exist.")

            for profile_name, build_type in conan_install_combinations:
                install_dir_name = self.install_combination_to_install_dir_name(profile_name, build_type)

                toolchains_manifest_path = fs.path.join(install_dir_name, get_toolchain_manifest_filename())
                with work_fs.open(toolchains_manifest_path, "r") as f:
                    manifest = json.load(f)

                for k, v in manifest.items():
                    if k not in self.toolchains_manifest.keys():
                        self.toolchains_manifest[k] = v

                third_party_manifest_path = fs.path.join(install_dir_name, get_third_party_manifest_filename())
                with work_fs.open(third_party_manifest_path, "r") as f:
                    manifest = json.load(f)

                toolchain_name = profile_name.split("-")[-1]  # TODO

                for k, v in manifest.items():
                    iface_key = base.build_target_key(k, toolchain_name, build_type)
                    assert (iface_key not in self.third_party_manifest.keys())
                    self.third_party_manifest[iface_key] = v

    def get_toolchains_manifest(self):
        return self.toolchains_manifest

    def get_third_party_manifest(self):
        return self.third_party_manifest



from __future__ import annotations

from conans import ConanFile, tools
from conans.model import Generator
from conans.errors import ConanInvalidConfiguration
import traceback
import jsz.rapidbuild.conan

required_conan_version = ">=1.43"


# noinspection PyBroadException
class RapidGenerator(Generator):
    # noinspection PyPropertyDefinition
    @property
    def filename(self):
        pass

    @property
    def content(self):
        try:
            self.conanfile.output.info("Rapid: Running...")

            rapid_os = jsz.rapidbuild.conan.to_rapid_os(self.conanfile.settings.os)
            if rapid_os is None:
                raise ConanInvalidConfiguration(f"Rapid: unsupported OS '{self.conanfile.settings.os}'.")

            rapid_toolchain = jsz.rapidbuild.conan.to_rapid_toolchain(self.conanfile.settings.compiler,
                                                                 self.conanfile.settings.compiler.version)
            if rapid_toolchain is None:
                raise ConanInvalidConfiguration(f"Rapid: unsupported compiler '{self.conanfile.settings.compiler}',"
                                                f" version '{self.conanfile.settings.compiler.version}'.")

            vc_vars = tools.vcvars_dict(self.conanfile)
            deps_cpp_info = self.conanfile.deps_cpp_info

            return {
                jsz.rapidbuild.conan.get_third_party_manifest_filename():
                    jsz.rapidbuild.conan.generate_third_party_manifest_from_conan_deps_info(deps_cpp_info),
                jsz.rapidbuild.conan.get_toolchain_manifest_filename():
                    jsz.rapidbuild.conan.generate_toolchain_manifest_from_conan_vc_vars(rapid_toolchain, vc_vars)
            }

        except Exception:
            traceback.print_exc()


class RapidGeneratorPackage(ConanFile):
    name = "rapidbuild"
    version = "0.1"
    license = "MIT"



from . import base


class ToolchainSettingsProviderDefault(base.ToolchainSettingsProvider):
    @staticmethod
    def _get_toolchain_settings_msvc():
        return {
            "compiler_args": {
                "all": [
                    # warning level 4, as errors, all
                    '/W4', '/WX', '/Wall',
                    # Info: the compiler performed inlining on the given function, although it was not marked
                    # for inlining.
                    '/wd4711',
                    # do not report warnings from third parties
                    '/experimental:external', '/external:anglebrackets', '/external:W0',
                    # specifies that are sources are C + +
                    '/TP',
                    # set standard
                    '/std:c++17', '/permissive-',
                    # enable string pooling
                    '/GF',
                    # disable RTTI
                    '/GR-',
                    # s: Enables standard C++ stack unwinding.catches only standard C++ exceptions when you
                    #    use catch(...) syntax.
                    # c: Compiler assumes that functions declared as extern "C" never throw a C + + exception.
                    '/EHsc',
                    # no acquire / release semantics for volatile vars
                    '/volatile:iso',
                    # require to always define the class before declaring a pointer-to-member
                    '/vmb',
                    # disable M$ C extensions
                    # '/Za ' # not recommended for C + + code
                ],
                base.BUILD_TYPE_DEBUG: [
                    '/MDd',
                    # disable optimization
                    '/Od',
                    # debug format: PDB
                    '/Zi',
                    # faster PDB generation
                    '/Zf',
                    # floats according to standard
                    '/fp:strict',
                    # enable buffer security checks
                    '/GS',
                    '/sdl',
                    # enable control flow guards
                    '/guard:cf',
                    # enable EH continuation metadata(must be also present in linker args)
                    '/guard:ehcont',
                    # enable all runtime checks.RTCc rejects conformant code, so it is not supported by
                    # the C++ Standard Library
                    '/RTCsu'
                ],
                base.BUILD_TYPE_RELEASE: [
                    '/MD '
                    # enable optimization
                    '/O2 '
                    # fast floats
                    '/fp:fast '
                    # disable buffer security checks
                    '/GS- '
                    # + '/sdl- ' // it overrides / GS -
                    # disable control flow guards
                    '/guard:cf- '
                    # disable EH continuation metadata
                    '/guard:ehcont- '
                ],
            },
            "definitions": {
                "all": {
                    "_MT": None,
                    "_DLL": None,
                },
                base.BUILD_TYPE_DEBUG: {
                    "_DEBUG": None,
                },
                base.BUILD_TYPE_RELEASE: {
                },
            },
            "linker_args": {
                "all": [
                    "/WX",
                    "/IGNORE:4099",
                    "/machine:x64"
                ],
                base.BUILD_TYPE_DEBUG: [
                    "/DEBUG",
                    "/GUARD:CF"
                ],
                base.BUILD_TYPE_RELEASE: [
                    "/GUARD:CF"
                ],
            },
            "link_libs": {
                "all": [
                    'kernel32.lib',
                    'user32.lib',
                    'gdi32.lib',
                    'shell32.lib',
                    'winspool.lib',
                    'ole32.lib',
                    'oleaut32.lib',
                    'uuid.lib',
                    'comdlg32.lib',
                    'advapi32.lib',
                ],
                base.BUILD_TYPE_DEBUG: [
                    'ucrtd.lib',
                    'vcruntimed.lib',
                    'msvcrtd.lib',
                    'msvcprtd.lib',
                ],
                base.BUILD_TYPE_RELEASE: [
                    'ucrt.lib',
                    'vcruntime.lib',
                    'msvcrt.lib',
                    'msvcprt.lib',
                ],
            },
        }

    @staticmethod
    def _get_toolchain_settings_llvm():
        return {
            "compiler_args": {
                "all": [
                    # warnings
                    '-Wall', '-Wextra', '-pedantic', '-pedantic-errors', '-Werror',
                    # set standard
                    '-std=c++17',
                    # disable RTTI
                    '-fno-rtti',
                    # disable exceptions
                    '-fno-exceptions',
                    # use a best-case representation method for member pointers
                    '-fcomplete-member-pointers',
                ],
                base.BUILD_TYPE_DEBUG: [
                    '-O0',
                    '-g',
                ],
                base.BUILD_TYPE_RELEASE: [
                    '-O3',
                    '-ffast-math',
                ],
            },
            "definitions": {
                "all": {
                    "_MT": None,
                    "_DLL": None,
                },
                base.BUILD_TYPE_DEBUG: {
                    "_DEBUG": None,
                },
                base.BUILD_TYPE_RELEASE: {
                },
            },
            "linker_args": {
                "all": [
                    "/WX",
                    "/machine:x64"
                ],
                base.BUILD_TYPE_DEBUG: [
                    "/DEBUG",
                    "/GUARD:CF"
                ],
                base.BUILD_TYPE_RELEASE: [
                    "/GUARD:CF"
                ],
            },
            "link_libs": {
                "all": [
                    'kernel32.lib',
                    'user32.lib',
                    'gdi32.lib',
                    'shell32.lib',
                    'winspool.lib',
                    'ole32.lib',
                    'oleaut32.lib',
                    'uuid.lib',
                    'comdlg32.lib',
                    'advapi32.lib',
                ],
                base.BUILD_TYPE_DEBUG: [
                    'ucrtd.lib',
                    'vcruntimed.lib',
                    'msvcrtd.lib',
                    'msvcprtd.lib',
                ],
                base.BUILD_TYPE_RELEASE: [
                    'ucrt.lib',
                    'vcruntime.lib',
                    'msvcrt.lib',
                    'msvcprt.lib',
                ],
            },
        }

    def get_toolchain_settings(self):
        return {
            base.TOOLCHAIN_MSVC16: self._get_toolchain_settings_msvc(),
            base.TOOLCHAIN_LLVM13: self._get_toolchain_settings_llvm(),
        }



#!/usr/bin/env
from dataclasses import dataclass
from typing import Any, Union
import yaml
import os.path
import fs.path
from fs.base import FS
from fs.osfs import OSFS
import fs.errors

from . import base
from .conan import SystemConfigManifestProviderConan
from .fastbuild import BuildScriptEmitterFastbuild
from .toolchain import ToolchainSettingsProviderDefault


class WorkspaceInvalidFormatError(Exception):
    pass


@dataclass
class TargetScopedProperties:
    definitions: dict[Any, Any]
    dependencies: list[str]
    externals: list[str]

    @staticmethod
    def from_dict(d: dict):
        if not isinstance(d, dict):
            raise WorkspaceInvalidFormatError
        if not set(d.keys()).issubset({"definitions", "dependencies", "externals"}):
            raise WorkspaceInvalidFormatError
        definitions = d.get("definitions", {})
        if not isinstance(definitions, dict):
            raise WorkspaceInvalidFormatError
        dependencies = d.get("dependencies", [])
        if not isinstance(dependencies, list):
            raise WorkspaceInvalidFormatError
        externals = d.get("externals", [])
        if not isinstance(externals, list):
            raise WorkspaceInvalidFormatError
        return TargetScopedProperties(definitions=definitions, dependencies=dependencies, externals=externals)


@dataclass
class TargetDefinition:
    name: str
    kind: base.TargetKind
    private_props: TargetScopedProperties
    public_props: TargetScopedProperties

    @staticmethod
    def from_dict(name: str, d: dict):
        if not isinstance(d, dict):
            raise WorkspaceInvalidFormatError
        if not set(d.keys()).issubset({"kind", "public", "private"}):
            raise WorkspaceInvalidFormatError
        if "kind" not in d.keys():
            raise WorkspaceInvalidFormatError
        try:
            kind = base.TargetKind.from_string(d["kind"])
        except KeyError:
            raise WorkspaceInvalidFormatError
        public_props = TargetScopedProperties.from_dict(d.get("public", {}))
        private_props = TargetScopedProperties.from_dict(d.get("private", {}))
        return TargetDefinition(name=name, kind=kind, private_props=private_props, public_props=public_props)


@dataclass
class WorkspaceDefinition:
    targets: list[TargetDefinition]

    @staticmethod
    def from_dict(d):
        if not isinstance(d, dict):
            raise WorkspaceInvalidFormatError
        if len(d) == 0:
            raise WorkspaceInvalidFormatError
        targets = [TargetDefinition.from_dict(name, v) for name, v in d.items()]
        return WorkspaceDefinition(targets=targets)


def append_unique(list_: list[Any], item: any):
    if item not in list_:
        list_.append(item)


def _validate_third_party_manifest(manifest):
    if not isinstance(manifest, dict):
        return False
    return True


def _validate_toolchains_manifest(manifest):
    if not isinstance(manifest, dict):
        return False
    if len(manifest.keys()) == 0:
        return False
    return True


def _create_targets_interfaces_and_implementations(source_dir_abs_path, toolchains_manifest, build_types,
                                                   third_party_manifest, workspace_def: WorkspaceDefinition):
    assert(len(source_dir_abs_path) != 0)
    assert(len(build_types) != 0)
    assert(_validate_toolchains_manifest(toolchains_manifest))
    assert(_validate_third_party_manifest(third_party_manifest))

    targets_interfaces = dict(third_party_manifest)
    targets_impls = {}

    for toolchain_name in toolchains_manifest.keys():
        for build_type in build_types:
            for target_def in workspace_def.targets:
                iface = {
                    "kind": "",
                    "definitions": {},
                    "include_dirs": [],
                    "link_libs": [],
                    "link_libs_external": [],
                    "link_libs_external_dirs": [],
                    "runtime_libs": []
                }

                impl = {
                    "kind": "",
                    "definitions": {},
                    "source_dir": "",
                    "include_dirs": [],
                    "link_libs": [],
                    "link_libs_external": [],
                    "link_libs_external_dirs": [],
                    "runtime_libs": []
                }

                target_public_include_dir_name = "include"
                target_private_include_dir_name = "src"
                target_source_dir_name = "src"

                target_public_include_path = fs.path.join(source_dir_abs_path, target_def.name,
                                                          target_public_include_dir_name)
                target_private_include_path = fs.path.join(source_dir_abs_path, target_def.name,
                                                           target_private_include_dir_name)
                target_source_dir_path = fs.path.join(source_dir_abs_path, target_def.name,
                                                      target_source_dir_name)

                iface["kind"] = target_def.kind
                impl["kind"] = target_def.kind

                iface["include_dirs"].append(target_public_include_path)
                impl["include_dirs"].append(target_public_include_path)
                impl["include_dirs"].append(target_private_include_path)
                impl["source_dir"] = target_source_dir_path

                iface["link_libs"].append(target_def.name)

                # process public section

                # Own public definitions -> interface definitions + impl definitions
                for name, value in target_def.public_props.definitions.items():
                    # TODO: resolve conflicts (A=3, A=4)
                    if name not in iface["definitions"]:
                        iface["definitions"][name] = value
                    if name not in impl["definitions"]:
                        impl["definitions"][name] = value

                for dependency in target_def.public_props.dependencies:
                    dep_iface = targets_interfaces[base.build_target_key(dependency, toolchain_name, build_type)]

                    if dep_iface["kind"] == base.TargetKind.EXECUTABLE:
                        raise RuntimeError("Cannot declare dep on an exe.")

                    for link_lib in dep_iface["link_libs"]:
                        append_unique(iface["link_libs"], link_lib)
                        append_unique(impl["link_libs"], link_lib)

                    for link_lib in dep_iface["link_libs_external"]:
                        append_unique(iface["link_libs_external"], link_lib)
                        append_unique(impl["link_libs_external"], link_lib)

                    for link_lib in dep_iface["link_libs_external_dirs"]:
                        append_unique(iface["link_libs_external_dirs"], link_lib)
                        append_unique(impl["link_libs_external_dirs"], link_lib)

                for dependency in target_def.public_props.externals:
                    dep_iface = targets_interfaces[
                        base.build_target_key(dependency, toolchain_name, build_type)]

                    for link_lib in dep_iface["link_libs_external"]:
                        append_unique(iface["link_libs_external"], link_lib)
                        append_unique(impl["link_libs_external"], link_lib)

                    for link_lib in dep_iface["link_libs_external_dirs"]:
                        append_unique(iface["link_libs_external_dirs"], link_lib)
                        append_unique(impl["link_libs_external_dirs"], link_lib)

                deps = target_def.public_props.dependencies + target_def.public_props.externals

                for dependency in deps:
                    dep_iface = targets_interfaces[base.build_target_key(dependency, toolchain_name, build_type)]

                    for name, value in dep_iface["definitions"].items():
                        # TODO: resolve conflicts (A=3, A=4)
                        if name not in iface["definitions"]:
                            iface["definitions"][name] = value
                        if name not in impl["definitions"]:
                            impl["definitions"][name] = value

                    for d in dep_iface["include_dirs"]:
                        append_unique(iface["include_dirs"], d)
                        append_unique(impl["include_dirs"], d)

                    for d in dep_iface["runtime_libs"]:
                        append_unique(iface["runtime_libs"], d)
                        # TODO: Is that a typo?
                        append_unique(impl["runtime_libs"], d)

                    if dep_iface["kind"] == base.TargetKind.DYNAMIC_LIB:
                        append_unique(iface["runtime_libs"], dependency)
                        append_unique(impl["runtime_libs"], dependency)

                # process private section

                for name, value in target_def.private_props.definitions.items():
                    # TODO: resolve conflicts (A=3, A=4)
                    if name not in impl["definitions"]:
                        impl["definitions"][name] = value

                for dependency in target_def.private_props.dependencies:
                    dep_iface = targets_interfaces[base.build_target_key(dependency, toolchain_name, build_type)]

                    if dep_iface["kind"] == base.TargetKind.EXECUTABLE:
                        raise RuntimeError("Cannot declare dep on an exe.")

                    for link_lib in dep_iface["link_libs"]:
                        append_unique(iface["link_libs"], link_lib)
                        append_unique(impl["link_libs"], link_lib)

                    for link_lib in dep_iface["link_libs_external"]:
                        append_unique(iface["link_libs_external"], link_lib)
                        append_unique(impl["link_libs_external"], link_lib)

                    for link_lib in dep_iface["link_libs_external_dirs"]:
                        append_unique(iface["link_libs_external_dirs"], link_lib)
                        append_unique(impl["link_libs_external_dirs"], link_lib)

                for dependency in target_def.private_props.externals:
                    dep_iface = targets_interfaces[base.build_target_key(dependency, toolchain_name, build_type)]

                    for link_lib in dep_iface["link_libs_external"]:
                        append_unique(iface["link_libs_external"], link_lib)
                        append_unique(impl["link_libs_external"], link_lib)

                    for link_lib in dep_iface["link_libs_external_dirs"]:
                        append_unique(iface["link_libs_external_dirs"], link_lib)
                        append_unique(impl["link_libs_external_dirs"], link_lib)

                deps = target_def.private_props.dependencies + target_def.private_props.externals

                for dependency in deps:
                    dep_iface = targets_interfaces[
                        base.build_target_key(dependency, toolchain_name, build_type)]

                    # print(json.dumps(dep_iface, indent=4))

                    for name, value in dep_iface["definitions"].items():
                        # TODO: resolve conflicts (A=3, A=4)
                        if name not in impl["definitions"]:
                            impl["definitions"][name] = value

                    for d in dep_iface["include_dirs"]:
                        append_unique(impl["include_dirs"], d)

                    for d in dep_iface["runtime_libs"]:
                        append_unique(iface["runtime_libs"], d)
                        append_unique(impl["runtime_libs"], d)

                    if dep_iface["kind"] == base.TargetKind.DYNAMIC_LIB:
                        append_unique(iface["runtime_libs"], dependency)
                        append_unique(impl["runtime_libs"], dependency)

                # save result
                target_key = base.build_target_key(target_def.name, toolchain_name, build_type)

                # Executables don't need an interface
                if target_def.kind != base.TargetKind.EXECUTABLE:
                    targets_interfaces[target_key] = iface
                targets_impls[target_key] = impl

    return targets_interfaces, targets_impls


class Workspace:
    def __init__(self, *,
                 wks_fs: FS,
                 logger: base.Logger,
                 system_config_provider: base.SystemConfigManifestProvider,
                 toolchain_settings_provider: base.ToolchainSettingsProvider,
                 build_script_emitter: base.BuildScriptEmitter
                 ):

        self.fs = wks_fs

        self.configure_dir_name = "configure"
        self.source_dir_name = "src"
        self.build_dir_name = "build"

        self.workspace_def: Union[WorkspaceDefinition, None] = None
        self.third_party_manifest: dict[str, Any] = {}
        self.toolchains_manifest: dict[str, Any] = {}

        self.build_types = [base.BUILD_TYPE_DEBUG, base.BUILD_TYPE_RELEASE]

        self.logger = logger
        self.system_config_provider = system_config_provider
        self.toolchain_settings_provider = toolchain_settings_provider
        self.build_script_emitter = build_script_emitter

    def __del__(self):
        self.fs.close()
        pass

    def _get_source_dir_path(self):
        return self.source_dir_name

    def _get_source_dir_abs_path(self):
        if self.fs.hassyspath(self.source_dir_name):
            return self.fs.getsyspath(self.source_dir_name)
        return f"wks://{self.source_dir_name}"

    def _get_configure_dir_abs_path(self):
        if self.fs.hassyspath(self.configure_dir_name):
            return self.fs.getsyspath(self.configure_dir_name)
        return f"wks://{self.configure_dir_name}"

    def _get_build_dir_abs_path(self):
        if self.fs.hassyspath(self.build_dir_name):
            return self.fs.getsyspath(self.build_dir_name)
        return f"wks://{self.build_dir_name}"

    def configure(self, *, quick_mode: bool) -> bool:

        #
        # Load workspace target definitions
        #
        targets_file_path = fs.path.join(self.source_dir_name, "rapid_targets.yml")
        self.logger.log_info(f"Reading target definitions '{targets_file_path}'...")

        try:
            with self.fs.open(targets_file_path, 'r') as f:
                targets_defs = yaml.load(f, yaml.Loader)
        except fs.errors.ResourceNotFound:
            self.logger.log_error(f"File '{targets_file_path}' does not exist.")
            return False

        try:
            self.workspace_def = WorkspaceDefinition.from_dict(targets_defs)
        except WorkspaceInvalidFormatError:
            self.logger.log_error(f"File '{targets_file_path}' is not a valid targets definition file.")
            return False

        self.logger.log_info("Workspace definition loaded.")

        # print(json.dumps(target_defs, indent=2))

        #
        # Prepare configuration directory, provide manifests
        #
        self.fs.makedir(self.configure_dir_name, recreate=True)

        self.system_config_provider.run(self._get_configure_dir_abs_path())

        self.third_party_manifest = self.system_config_provider.get_third_party_manifest()

        if not _validate_third_party_manifest(self.third_party_manifest):
            self.logger.log_error("Provided third party manifest is invalid.")
            return False

        self.toolchains_manifest = self.system_config_provider.get_toolchains_manifest()

        if not _validate_toolchains_manifest(self.toolchains_manifest):
            self.logger.log_error("Provided toolchains manifest is invalid.")
            return False

        #
        # Build targets implementations
        #
        targets_interfaces, targets_impls = _create_targets_interfaces_and_implementations(
            self._get_source_dir_abs_path(),
            self.toolchains_manifest,
            self.build_types,
            self.third_party_manifest,
            self.workspace_def
        )

        assert(isinstance(targets_interfaces, dict))
        assert(isinstance(targets_impls, dict) and len(targets_impls.keys()) != 0)

        #
        # Generate build script for targets based on implementations
        #

        build_script_filename = self.build_script_emitter.filename()
        build_script_contents = self.build_script_emitter.contents(
            self._get_build_dir_abs_path(),
            self.toolchains_manifest,
            self.toolchain_settings_provider.get_toolchain_settings(),
            self.build_types,
            [d.name for d in self.workspace_def.targets],
            targets_impls
        )

        build_script_path = fs.path.join(self.configure_dir_name, build_script_filename)

        self.logger.log_info(f"Writing build script '{build_script_path}'...")
        with self.fs.open(build_script_path, 'w') as f:
            f.write(build_script_contents)

        self.logger.log_info("Configuring done.")
        return True


def open_workspace(wks_dir: str) -> Workspace:

    wks_fs = OSFS(os.path.abspath(wks_dir))

    logger = base.LoggerDefault()

    system_config_provider = SystemConfigManifestProviderConan(
        logger=logger,
        process_runner=base.ProcessRunnerDefault(),
        conanfile_path=wks_fs.getsyspath("conanfile.py"),
        execute_conan=True,
        base_to_conan_build_type={
            base.BUILD_TYPE_DEBUG: "Debug",
            base.BUILD_TYPE_RELEASE: "Release"
        }
    )

    wks = Workspace(
        wks_fs=wks_fs,
        logger=logger,
        system_config_provider=system_config_provider,
        toolchain_settings_provider=ToolchainSettingsProviderDefault(),
        build_script_emitter=BuildScriptEmitterFastbuild()
    )
    return wks



from contextlib import contextmanager
from typing import Any
import fs

from . import base


class BFFEmitter:
    GENERATED_FILE_HEADER = [
        "/////////////////////////////////////////////////////////////",
        "// This is a generated file, all manual changes will be lost!",
        "/////////////////////////////////////////////////////////////",
    ]
    INDENT_SIZE = 4

    def __init__(self):
        self._lines = []
        self._indent_level = 0

        for line in self.GENERATED_FILE_HEADER:
            self._emit_line(line)

    def _indent(self):
        self._indent_level += 1

    def _dedent(self):
        assert self._indent_level > 0
        self._indent_level -= 1

    def _emit_line(self, content: str):
        if len(content) != 0:
            self._lines.append("{}{}".format(' ' * self._indent_level * self.INDENT_SIZE, content))
        else:
            self._lines.append("")

    def _emit_array(self, name: str, values: list):
        self._emit_line(".{} = ".format(name))
        self._emit_line("{")
        self._indent()
        for v in values:
            self._emit_line("{},".format(self._format_value(v, can_split_str=False)))
        self._dedent()
        self._emit_line("}")

    def _emit_struct(self, name: str, dictionary: dict):
        self._emit_line(".{} = ".format(name))
        self._emit_line("[")
        self._indent()
        for key, value in dictionary.items():
            self.emit_var_assignment(key, value)
        self._dedent()
        self._emit_line("]")

    def _begin_function(self, fn_name: str, args: str):
        self._emit_line("{}('{}')".format(fn_name, args))
        self._emit_line("{")
        self._indent()

    def _end_function(self):
        self._dedent()
        self._emit_line("}")
        self._emit_line("")

    def once(self):
        self._emit_line("#once")

    def include(self, path: str):
        self._emit_line("#include \"{}\"".format(path))

    def line_break(self):
        self._emit_line("")

    def emit_var_assignment(self, name: str, value, *, should_split_str=False):
        if type(value) == list:
            self._emit_array(name, value)
        elif type(value) == dict:
            self._emit_struct(name, value)
        else:
            self._emit_line(".{} = {}".format(name, self._format_value(value, can_split_str=should_split_str)))

    @contextmanager
    def library(self, name: str):
        self._begin_function("Library", name)
        yield
        self._end_function()

    @contextmanager
    def object_list(self, name: str):
        self._begin_function("ObjectList", name)
        yield
        self._end_function()

    @contextmanager
    def dll(self, name: str):
        self._begin_function("DLL", name)
        yield
        self._end_function()

    @contextmanager
    def exe(self, name: str):
        self._begin_function("Executable", name)
        yield
        self._end_function()

    @contextmanager
    def alias(self, name: str):
        self._begin_function("Alias", name)
        yield
        self._end_function()

    @contextmanager
    def copy_dir(self, name: str):
        self._begin_function("CopyDir", name)
        yield
        self._end_function()

    @contextmanager
    def compiler(self, name: str):
        self._begin_function("Compiler", name)
        yield
        self._end_function()

    def build(self):
        return "\n".join(self._lines)

    def _format_value(self, value, *, can_split_str):

        def format_default(v):
            return "{}".format(v)

        def str_smart_split(s: str):
            splits = []
            inside_quoted = False
            delim = " "
            curr = ""
            for c in s:
                if not inside_quoted:
                    if c == "\"":
                        inside_quoted = True
                        curr += c
                    elif c == delim:
                        splits.append(curr)
                        curr = delim
                    else:
                        curr += c
                else:
                    if c == "\"":
                        inside_quoted = False
                        curr += c
                    else:
                        curr += c
            if len(curr) != 0:
                splits.append(curr)
            return splits

        def format_string(s: str):
            if len(s) == 0:
                return "''"

            if not can_split_str:
                return f"'{s}'"

            lines = []
            col_limit = 80

            if len(s) > col_limit:
                lines = str_smart_split(s)
            else:
                lines.append(s)

            result = ""
            result += "'{}'".format(lines[0])
            if len(lines) > 1:
                for line in lines[1:]:
                    result += f"\n{' ' * (self._indent_level + 1) * self.INDENT_SIZE}+ '{line}'"
            return result

        def format_bool(b: bool):
            return "true" if b else "false"

        formatters = {
            str: format_string,
            bool: format_bool,
        }

        f = formatters.get(type(value), format_default)
        return f(value)


def format_compiler_node_name(toolchain_name: str) -> str:
    return f"compiler-{toolchain_name}"


def format_target_node_name(target_name: str, toolchain_name: str, build_type_name: str) -> str:
    return f"{target_name}-{toolchain_name}-{build_type_name}"


def format_definition_compiler_arg_msvc(name: str, value: Any) -> str:
    if value is not None:
        return f"/D\"{name}={value}\""
    else:
        return f"/D\"{name}\""


def format_include_dir_compiler_arg_msvc(include_dir: str) -> str:
    return f"/I\"{include_dir}\""


def format_definition_compiler_arg_llvm(name: str, value: Any) -> str:
    if value is not None:
        return f"-D\"{name}={value}\""
    else:
        return f"-D\"{name}\""


def format_include_dir_compiler_arg_llvm(include_dir: str) -> str:
    return f"-I\"{include_dir}\""


def format_lib_dir_linker_arg_msvc(include_dir: str) -> str:
    return f"/libpath:\"{include_dir}\""


def format_lib_dir_linker_arg_llvm(include_dir: str) -> str:
    return f"/libpath:\"{include_dir}\""


class BuildScriptEmitterFastbuild(base.BuildScriptEmitter):
    ALL_TARGETS_TARGET_NAME = "rapid_all"

    def filename(self):
        return "fbuild.bff"

    def contents(self, build_dir_abs_path, toolchains_manifest, toolchains_settings, build_types, targets_names,
                 targets_impls) -> str:

        intermediate_dir_abs_path = fs.path.join(build_dir_abs_path, "intermediate")
        deploy_dir_abs_path = fs.path.join(build_dir_abs_path, "bin")

        out_bff = BFFEmitter()
        out_bff.once()

        for toolchain_name, toolchain_def in toolchains_manifest.items():
            with out_bff.compiler(format_compiler_node_name(toolchain_name)):
                out_bff.emit_var_assignment("Executable", toolchain_def["compiler_path"])
                out_bff.emit_var_assignment("CompilerFamily", toolchain_def["compiler_family"])
                out_bff.emit_var_assignment("ExtraFiles", toolchain_def["compiler_extra_files"])

        include_dir_compiler_arg_formatters = {
            base.TOOLCHAIN_MSVC16: format_include_dir_compiler_arg_msvc,
            base.TOOLCHAIN_LLVM13: format_include_dir_compiler_arg_llvm,
        }

        preprocessor_definition_compiler_arg_formatters = {
            base.TOOLCHAIN_MSVC16: format_definition_compiler_arg_msvc,
            base.TOOLCHAIN_LLVM13: format_definition_compiler_arg_llvm,
        }

        lib_dir_linker_arg_formatters = {
            base.TOOLCHAIN_MSVC16: format_lib_dir_linker_arg_msvc,
            base.TOOLCHAIN_LLVM13: format_lib_dir_linker_arg_llvm,
        }

        all_deploy_nodes: list[str] = []
        per_toolchain_deploy_nodes: dict[str, list[str]] = {toolchain_name: [] for toolchain_name in
                                                            toolchains_manifest.keys()}
        per_build_type_deploy_nodes: dict[str, list[str]] = {build_type: [] for build_type in build_types}

        for toolchain_name, toolchain_def in toolchains_manifest.items():

            format_definition_compiler_arg = preprocessor_definition_compiler_arg_formatters[toolchain_name]
            format_include_dir_compiler_arg = include_dir_compiler_arg_formatters[toolchain_name]
            format_lib_dir_linker_arg = lib_dir_linker_arg_formatters[toolchain_name]

            toolchain_include_dirs = toolchain_def["toolchain_include_dirs"]
            toolchain_lib_dirs = toolchain_def["toolchain_lib_dirs"]

            current_toolchain_settings = toolchains_settings[toolchain_name]
            toolchain_definitions = current_toolchain_settings["definitions"]
            toolchain_compiler_args = current_toolchain_settings["compiler_args"]
            toolchain_linker_args = current_toolchain_settings["linker_args"]
            toolchain_link_libs = current_toolchain_settings["link_libs"]

            for build_type in build_types:

                toolchain_build_type_definitions = toolchain_definitions["all"] | toolchain_definitions[build_type]
                toolchain_build_type_compiler_args = toolchain_compiler_args["all"] + toolchain_compiler_args[
                    build_type]
                toolchain_build_type_linker_args = toolchain_linker_args["all"] + toolchain_linker_args[build_type]
                toolchain_build_type_link_libs = toolchain_link_libs["all"] + toolchain_link_libs[build_type]

                for target_name in targets_names:

                    target_impl = targets_impls[base.build_target_key(target_name, toolchain_name, build_type)]
                    target_node_name = format_target_node_name(target_name, toolchain_name, build_type)

                    target_kind = target_impl["kind"]

                    target_source_dir_path = target_impl["source_dir"]
                    target_source_dir_glob_pattern = ["*.c", "*.cpp", "*.cxx"]
                    target_output_dir_path = fs.path.join(intermediate_dir_abs_path, target_node_name)

                    def build_compiler_options():
                        compiler_options = []

                        if toolchain_name == base.TOOLCHAIN_MSVC16:
                            intermediate_pdb_path = fs.path.join(target_output_dir_path, f"{target_node_name}-lib.pdb")
                            compiler_options += ["/nologo", "/c", "\"%1\"", "/Fo\"%2\"",
                                                 f"/Fd\"{intermediate_pdb_path}\""]
                        elif toolchain_name == base.TOOLCHAIN_LLVM13:
                            compiler_options += ["-c", "%1", "-o", "%2"]
                        else:
                            raise RuntimeError("unsupported toolchain")

                        compiler_options += toolchain_build_type_compiler_args
                        compiler_options += [format_include_dir_compiler_arg(include_dir)
                                             for include_dir in toolchain_include_dirs]
                        compiler_options += [format_definition_compiler_arg(name, value)
                                             for name, value in toolchain_build_type_definitions.items()]

                        compiler_options += [format_include_dir_compiler_arg(include_dir)
                                             for include_dir in target_impl["include_dirs"]]
                        compiler_options += [format_definition_compiler_arg(name, value)
                                             for name, value in target_impl["definitions"].items()]
                        return compiler_options

                    def build_linker_options(*, dll=False):
                        linker_options = []

                        output_pdb_path = fs.path.join(target_output_dir_path, f"{target_node_name}.pdb")

                        if toolchain_name == base.TOOLCHAIN_MSVC16:
                            linker_options += ["/nologo", "\"%1\"", "/out:\"%2\"", f"/pdb:\"{output_pdb_path}\""]
                            if dll:
                                linker_options += ["/dll"]
                        elif toolchain_name == base.TOOLCHAIN_LLVM13:
                            linker_options += ["/nologo", "\"%1\"", "/out:\"%2\"", f"/pdb:\"{output_pdb_path}\""]
                            if dll:
                                linker_options += ["/dll"]
                        else:
                            raise RuntimeError("unsupported toolchain")

                        linker_options += toolchain_build_type_linker_args

                        linker_options += toolchain_build_type_link_libs
                        linker_options += [format_lib_dir_linker_arg(lib_dir)
                                           for lib_dir in toolchain_lib_dirs]

                        # External libs are added manully, not through "Libraries" variable.
                        # If the extension is skipped, fastbuild tries to use "obj"
                        linker_options += [f"{lib}.lib"
                                           for lib in target_impl["link_libs_external"]]
                        linker_options += [format_lib_dir_linker_arg(lib_dir)
                                           for lib_dir in target_impl["link_libs_external_dirs"]]

                        return linker_options

                    def build_librarian_options():
                        librarian_options = []

                        if toolchain_name == base.TOOLCHAIN_MSVC16:
                            librarian_options += ["/NOLOGO", "/MACHINE:X64", "/WX", "\"%1\"", "/OUT:\"%2\""]
                        elif toolchain_name == base.TOOLCHAIN_LLVM13:
                            librarian_options += ["rc", "%2", "%1"]
                        else:
                            raise RuntimeError("unsupported toolchain")

                        return librarian_options

                    if target_kind == base.TargetKind.STATIC_LIB:

                        compiler_options = build_compiler_options()
                        compiler_input_path = target_source_dir_path
                        compiler_input_pattern = target_source_dir_glob_pattern
                        compiler_output_path = target_output_dir_path

                        librarian_options = build_librarian_options()
                        librarian_output_path = fs.path.join(target_output_dir_path, f"{target_node_name}.lib")

                        with out_bff.library(target_node_name):
                            out_bff.emit_var_assignment("Compiler", format_compiler_node_name(toolchain_name))
                            out_bff.emit_var_assignment("CompilerInputPath", compiler_input_path)
                            out_bff.emit_var_assignment("CompilerInputPathRecurse", True)
                            out_bff.emit_var_assignment("CompilerInputPattern", compiler_input_pattern)
                            out_bff.emit_var_assignment("CompilerOutputPath", compiler_output_path)
                            out_bff.emit_var_assignment("CompilerOptions", " ".join(compiler_options),
                                                        should_split_str=True)
                            out_bff.line_break()
                            out_bff.emit_var_assignment("Librarian", toolchain_def["librarian_path"])
                            out_bff.emit_var_assignment("LibrarianOutput", librarian_output_path)
                            out_bff.emit_var_assignment("LibrarianOptions", " ".join(librarian_options),
                                                        should_split_str=True)

                        deploy_target_name = f"deploy-{target_node_name}"
                        with out_bff.alias(deploy_target_name):
                            out_bff.emit_var_assignment("Targets", target_node_name)

                        all_deploy_nodes.append(deploy_target_name)
                        per_build_type_deploy_nodes[build_type].append(deploy_target_name)
                        per_toolchain_deploy_nodes[toolchain_name].append(deploy_target_name)

                    elif target_kind == base.TargetKind.DYNAMIC_LIB:

                        compiler_options = build_compiler_options()
                        compiler_input_path = target_source_dir_path
                        compiler_input_pattern = target_source_dir_glob_pattern
                        compiler_output_path = target_output_dir_path

                        object_list_node_name = f"obj-{target_node_name}"

                        with out_bff.object_list(object_list_node_name):
                            out_bff.emit_var_assignment("Hidden", True)
                            out_bff.emit_var_assignment("Compiler", format_compiler_node_name(toolchain_name))
                            out_bff.emit_var_assignment("CompilerInputPath", compiler_input_path)
                            out_bff.emit_var_assignment("CompilerInputPathRecurse", True)
                            out_bff.emit_var_assignment("CompilerInputPattern", compiler_input_pattern)
                            out_bff.emit_var_assignment("CompilerOutputPath", compiler_output_path)
                            out_bff.emit_var_assignment("CompilerOptions", " ".join(compiler_options),
                                                        should_split_str=True)

                        #
                        linker_output_path = fs.path.join(target_output_dir_path, f"{target_node_name}.dll")

                        libraries = [object_list_node_name]
                        libraries += [format_target_node_name(lib, toolchain_name, build_type)
                                      for lib in target_impl["link_libs"]]

                        linker_options = build_linker_options(dll=True)

                        with out_bff.dll(target_node_name):
                            out_bff.emit_var_assignment("Libraries", libraries)
                            out_bff.emit_var_assignment("Linker", toolchain_def["linker_path"])
                            out_bff.emit_var_assignment("LinkerOutput", linker_output_path)
                            out_bff.emit_var_assignment("LinkerOptions", " ".join(linker_options),
                                                        should_split_str=True)

                        #
                        prebuild_deps = [target_node_name]
                        prebuild_deps += [f"deploy-{format_target_node_name(lib, toolchain_name, build_type)}"
                                          for lib in target_impl["runtime_libs"]]

                        deploy_target_name = f"deploy-{target_node_name}"
                        with out_bff.copy_dir(deploy_target_name):
                            out_bff.emit_var_assignment("PreBuildDependencies", prebuild_deps)
                            out_bff.emit_var_assignment("SourcePaths", target_output_dir_path)
                            out_bff.emit_var_assignment("SourcePathsPattern", ["*.dll", "*.pdb"])
                            out_bff.emit_var_assignment("Dest", deploy_dir_abs_path)

                        all_deploy_nodes.append(deploy_target_name)
                        per_build_type_deploy_nodes[build_type].append(deploy_target_name)
                        per_toolchain_deploy_nodes[toolchain_name].append(deploy_target_name)

                    elif target_kind == base.TargetKind.EXECUTABLE:

                        compiler_options = build_compiler_options()
                        compiler_input_path = target_source_dir_path
                        compiler_input_pattern = target_source_dir_glob_pattern
                        compiler_output_path = target_output_dir_path

                        object_list_node_name = f"obj-{target_node_name}"

                        with out_bff.object_list(object_list_node_name):
                            out_bff.emit_var_assignment("Hidden", True)
                            out_bff.emit_var_assignment("Compiler", format_compiler_node_name(toolchain_name))
                            out_bff.emit_var_assignment("CompilerInputPath", compiler_input_path)
                            out_bff.emit_var_assignment("CompilerInputPathRecurse", True)
                            out_bff.emit_var_assignment("CompilerInputPattern", compiler_input_pattern)
                            out_bff.emit_var_assignment("CompilerOutputPath", compiler_output_path)
                            out_bff.emit_var_assignment("CompilerOptions", " ".join(compiler_options),
                                                        should_split_str=True)

                        #
                        linker_output_path = fs.path.join(target_output_dir_path, f"{target_node_name}.exe")

                        libraries = [object_list_node_name]
                        libraries += [format_target_node_name(lib, toolchain_name, build_type)
                                      for lib in target_impl["link_libs"]]

                        linker_options = build_linker_options(dll=False)

                        with out_bff.exe(target_node_name):
                            out_bff.emit_var_assignment("Libraries", libraries)
                            out_bff.emit_var_assignment("Linker", toolchain_def["linker_path"])
                            out_bff.emit_var_assignment("LinkerOutput", linker_output_path)
                            out_bff.emit_var_assignment("LinkerOptions", " ".join(linker_options),
                                                        should_split_str=True)

                        #
                        prebuild_deps = [target_node_name]
                        prebuild_deps += [f"deploy-{format_target_node_name(lib, toolchain_name, build_type)}"
                                          for lib in target_impl["runtime_libs"]]

                        deploy_target_name = f"deploy-{target_node_name}"
                        with out_bff.copy_dir(deploy_target_name):
                            out_bff.emit_var_assignment("PreBuildDependencies", prebuild_deps)
                            out_bff.emit_var_assignment("SourcePaths", target_output_dir_path)
                            out_bff.emit_var_assignment("SourcePathsPattern", ["*.exe", "*.pdb"])
                            out_bff.emit_var_assignment("Dest", deploy_dir_abs_path)

                        all_deploy_nodes.append(deploy_target_name)
                        per_build_type_deploy_nodes[build_type].append(deploy_target_name)
                        per_toolchain_deploy_nodes[toolchain_name].append(deploy_target_name)

                    else:
                        raise RuntimeError("unsupported target kind")

        for build_type, deploy_nodes in per_build_type_deploy_nodes.items():
            build_type_all_targets_node_name = f"{self.ALL_TARGETS_TARGET_NAME}-{build_type}"

            with out_bff.alias(build_type_all_targets_node_name):
                out_bff.emit_var_assignment("Hidden", True)
                out_bff.emit_var_assignment("Targets", deploy_nodes)

            with out_bff.alias(f"deploy-{build_type_all_targets_node_name}"):
                out_bff.emit_var_assignment("Targets", build_type_all_targets_node_name)

        for toolchain_name, deploy_nodes in per_toolchain_deploy_nodes.items():
            toolchain_all_targets_node_name = f"{self.ALL_TARGETS_TARGET_NAME}-{toolchain_name}"

            with out_bff.alias(toolchain_all_targets_node_name):
                out_bff.emit_var_assignment("Hidden", True)
                out_bff.emit_var_assignment("Targets", deploy_nodes)

            with out_bff.alias(f"deploy-{toolchain_all_targets_node_name}"):
                out_bff.emit_var_assignment("Targets", toolchain_all_targets_node_name)

        all_targets_node_name = f"{self.ALL_TARGETS_TARGET_NAME}"

        with out_bff.alias(all_targets_node_name):
            out_bff.emit_var_assignment("Hidden", True)
            out_bff.emit_var_assignment("Targets", all_deploy_nodes)

        with out_bff.alias(f"deploy-{all_targets_node_name}"):
            out_bff.emit_var_assignment("Targets", all_targets_node_name)

        return out_bff.build()
